/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2006  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*
**  File: TA3D_Exception.h
** Notes:
**   Cire: This file should be used to handle errors.  It is rather simple to use and
**           provides methods for tossing an exception and figuring out where it came
**           from, as well as who it went through.
**         To guard a block of code, with the error to be caught by anybody higher-up:
**           guard (Action);
**              ...
**           unguard;
**         If something happens between the guard and unguard, someone will eventually
**           catch it and display an error, or act on it.
**
**         To catch errors from lower blocks of code, do as above but put catches in
**            for uchar*, cException*, and any other types (including ...) that you
**            need.
*/
#pragma once

namespace TA3D
{
   namespace EXCEPTION
   {
      // prototypes: internal use only.
      extern void      IncrementIndent(void);
      extern void      DecrementIndent(void);
      extern uint32   GetIndentLevel(void);

      // Use this to throw a nice, more useful debugging string
      // Be sure to SetExceptionStatus (false) if you handle an exception.
      // Don't worry about manually setting it to true though.
#ifdef TA3D_PLATFORM_WINDOWS
      extern void __cdecl ThrowFatal( const char *Format, ... );
#else
      extern void ThrowFatal( const char *Format, ... );
#endif

      extern  bool IsExceptionInProgress (void);
      extern  void SetExceptionStatus (bool E);

      #ifdef DEBUG
		#ifdef TA3D_PLATFORM_WINDOWS
        	extern void __cdecl GlobalDebuggerPrintf( const char *Format, ... );
       	#else
        	extern void GlobalDebuggerPrintf( const char *Format, ... );
       	#endif
         #define dprintf GlobalDebuggerPrintf
      #else
         #define dprintf //
      #endif

      // In debug builds, let unhandled errors (referencing bad pointer, etc.) crash
      // and give us debugger control. Otherwise, for release builds, give us a
      // stack trace. This is used in the unguard macro below.
      #ifdef DEBUG
         #define UNKNOWN_ERROR_CATCH
      #else
         #define UNKNOWN_ERROR_CATCH \
               catch (...) \
               { \
                  DecrementIndent (); \
                  dprintf ("Tossing Unhandled Exception (%s)\n", __GUARD__BLOCK__NAME__); \
                  SetExceptionStatus (true); \
                  throw (new cException (__GUARD__BLOCK__NAME__, "Unhandled Exception"));\
               }
      #endif

      // Used for both debug and release builds
      #define guard(name) \
         { \
            static char *__GUARD__BLOCK__NAME__ = #name; \
            dprintf ("Guard:   %s\n", __GUARD__BLOCK__NAME__); \
            IncrementIndent (); \
            try \
            {


      // Don't need to use this if using unguard. Use this if using
      // guardcatch, et. al.
      #define guardend \
            DecrementIndent (); \
            dprintf ("Unguard: %s\n", __GUARD__BLOCK__NAME__); \
         }


#define unguard \
        } \
        catch (cException *E) \
        { \
            E->AddTosser (__GUARD__BLOCK__NAME__); \
            DecrementIndent (); \
            dprintf ("Tossing exception higher (%s)\n", __GUARD__BLOCK__NAME__); \
            SetExceptionStatus (true); \
            throw (E); \
        } \
        catch (string ErrorName) \
        { \
            DecrementIndent (); \
            dprintf ("Tossing string higher as exception (%s)\n", __GUARD__BLOCK__NAME__); \
            SetExceptionStatus (true); \
            throw (new cException (__GUARD__BLOCK__NAME__, ErrorName)); \
        } \
        UNKNOWN_ERROR_CATCH \
    guardend;


   // Use if you want to handle ExceptionD yourself
   #define guardcatch(varname) \
        } \
        catch (cException *varname) \

   #ifdef TA3D_SUPERGUARD
      #define sguard(a) guard(a)
      #define sunguard unguard
   #else
      #define sguard(a)
      #define sunguard
   #endif

   // Convenience macro
   #define SAFETRY(cmd) { guard (cmd); cmd; unguard; }


   // Will eat any exceptions caused by a piece of code
   #define IMPUNITY(cmd) { try { cmd; } catch (...) { } }

   // Disable exception handling and save executable size (and speed?)
   #ifdef TA3D_NOGUARD
      #undef guard
      #define guard(x)

      #undef unguard
      #define unguard

      #undef guardend
      #define guardend

      #undef guardcatch
      #define guardcatch(varname) cException *varname = NULL; if (false)
   #endif // TA3D_NOGUARD



      // The actual exception class
      class cException
      {
      public:
         cException( const String _Instigator, const String _ErrorName )
         {
            Instigator = _Instigator;
            ErrorName = _ErrorName;
            TossList = "";
            return;
         }

         void AddTosser( const String Tosser )
         {
            TossList += " <- " + Tosser;
            return;
         }

         const String GetInstigator (void)  { return (Instigator); }
         const String GetTossList   (void)  { return (TossList);   }
         const String GetErrorName  (void)  { return (ErrorName);  }

      private:
         String Instigator;
         String TossList;
         String ErrorName;
      }; // class cException
   } // namespace EXCEPTION
} // namespace TA3D 
